#tool
extends Control

signal divisions_counted(divisions)

export(Resource) var feedback_effect setget set_feedback_effect

export var min_value := 0.0
export var max_value := 1.0

export var horizontal_texts := 9
export var vertical_setions := 5
export var weak_right_color := Color.cadetblue
export var strong_left_color := Color.coral
export var selection_color := Color.bisque

export var length_scale := false

var last_left_click := 0
var multiselection_points := {} # of Curve: {index: nil}
var current_selection := {} # of Curve: {index: nil; vec2, when moving}
var is_selecting := false
var is_moving := false
var selection_rect := Rect2()
var selection_start_position := Vector2()


func _ready() -> void:
	multiselection_points[feedback_effect.weak_right_motor] = {}
	multiselection_points[feedback_effect.strong_left_motor] = {}
	clear_selection()


func _gui_input(event: InputEvent) -> void:
	var mb := event as InputEventMouseButton
	if mb:
		if mb.button_index == BUTTON_RIGHT and !mb.pressed:
			ensure_selection(mb.position)
			selection_switch_curve()
		elif mb.button_index == BUTTON_MIDDLE and !mb.pressed:
			ensure_selection(mb.position)
			remove_selection()
		elif mb.button_index == BUTTON_LEFT and mb.pressed and !mb.is_echo():
			if mb.shift:
				start_select(mb.position)
			elif selection_rect.has_point(mb.position):
				is_moving = true
				selection_start_position = mb.position
				for curve in current_selection:
					for point in current_selection[curve]:
						current_selection[curve][point] = curve.get_point_position(point)

			elif Time.get_ticks_msec() - last_left_click < 500:
				add_point(mb.position)
				select_closest_point_on_both_curves(mb.position)
				last_left_click -= 500
				is_moving = true
				selection_start_position = mb.position
			else:
				clear_selection()
				select_closest_point_on_both_curves(mb.position)
				if selection_empty():
					start_select(mb.position)
				else:
					selection_start_position = mb.position
					is_moving = true
		elif mb.button_index == BUTTON_LEFT and !mb.pressed:
			if Time.get_ticks_msec() - last_left_click < 500:
				add_point(mb.position)
				select_closest_point_on_both_curves(mb.position)

			if (!is_selecting and !is_moving) or (is_selecting and selection_empty()):
				clear_selection()
				select_closest_point_on_both_curves(mb.position)
				last_left_click = Time.get_ticks_msec()

			if is_selecting:
				for curve in multiselection_points:
					for point in multiselection_points[curve]:
						if !multiselection_points[curve][point]:
							continue
						if current_selection[curve].has(point):
							current_selection[curve].erase(point)
							continue
						current_selection[curve][point] = curve.get_point_position(point)

				calculate_selection_rect()

			is_selecting = false
			is_moving = false

	var mm := event as InputEventMouseMotion
	if mm:
		if is_selecting:
			selection_rect.size += mm.relative
			select_points_in_rect(selection_rect.abs(), feedback_effect.weak_right_motor, 1)
			select_points_in_rect(selection_rect.abs(), feedback_effect.strong_left_motor, -1)
		elif is_moving:
			_move_selection(sign(mm.relative.y) >= 0, calculate_curve_offset(mm.position))
	update()


func _move_selection(counting_up: bool, offset: Vector2) -> void:
	var new_selection := {feedback_effect.weak_right_motor: {}, feedback_effect.strong_left_motor: {}}
	for curve in current_selection:
		var points: Array = current_selection[curve].keys()
		points.sort()
		if counting_up:
			for point in points:
				new_selection[curve][_move_point(curve, point, offset)] = \
						current_selection[curve][point]
		else:
			for i in range(points.size(), 0, -1):
				new_selection[curve][_move_point(curve, points[i - 1], offset)] = \
						current_selection[curve][points[i - 1]]

	current_selection = new_selection
	calculate_selection_rect()


func _move_point(curve: Curve, point: int, offset: Vector2) -> int:
	var position := vec2_clamp(current_selection[curve][point] + offset)
	curve.set_point_value(point, position.y)
	return curve.set_point_offset(point, position.x)

func _draw() -> void:
	draw_rect(Rect2(Vector2(), rect_size), Color(0.06, 0.06, 0.06))

	var points := PoolVector2Array()
	var horizontal_sections := int(round(Engine.target_fps * feedback_effect.length))
	var horizontal_gap := rect_size.y / (horizontal_sections - 1)
	# Maybe force a horizontal line reduction if the gap is to small
	var divisions := 0
	while horizontal_gap < 16:
		divisions += 1
		horizontal_sections /= 2
		horizontal_gap = rect_size.y / (horizontal_sections - 1)
		if horizontal_sections <= 2:
			break
	emit_signal("divisions_counted", divisions)
	for i in range(horizontal_sections):
		points.push_back(Vector2(0, i * horizontal_gap))
		points.push_back(Vector2(rect_size.x, i * horizontal_gap))

	var vertical_gap := rect_size.x / (vertical_setions - 1)
	for i in range(vertical_setions):
		points.push_back(Vector2(i * vertical_gap, 0))
		points.push_back(Vector2(i * vertical_gap, rect_size.y))

	draw_multiline(points, Color(0.12, 0.12, 0.12))

	var font := get_font("font")
	var texts := horizontal_sections
	var text_gap := horizontal_gap
	while text_gap < font.get_height() * 2:
		texts /= 2
		text_gap = rect_size.y / (texts - 1)
		if texts <= 2:
			break
	for i in range(0, texts):
		var number: float = i * feedback_effect.length / (texts - 1)
		var position := Vector2(rect_size.x / 2, i * text_gap) \
				+ font.get_char_tx_size(ord("1")) * Vector2(0, 1) + Vector2(2 - 11, 4)
		draw_string(font, position, "%.2f" % number)

	_draw_curve_polygon(feedback_effect.weak_right_motor, weak_right_color, 1)
	_draw_curve_polygon(feedback_effect.strong_left_motor, strong_left_color, -1)

	if is_selecting or true:
		var rect := selection_rect.abs()
		draw_rect(rect, selection_color)
		var color := selection_color.darkened(0.2)
		color.a = 1
		draw_rect(rect, color, false)

	_draw_curve_handles(feedback_effect.weak_right_motor, weak_right_color, 1)
	_draw_curve_handles(feedback_effect.strong_left_motor, strong_left_color, -1)


func _draw_curve_handles(curve: Curve, color: Color, side := 1) -> void:
	color.a = 1
	for i in curve.get_point_count():
		var position := curve_to_canvas(curve.get_point_position(i), side)
		var handle_color := color
		if (get_local_mouse_position() - position).length_squared() < 64:
			handle_color = color.lightened(0.3)
		if i in current_selection[curve]:
			handle_color = handle_color.darkened(0.5)
		draw_circle(position, 3, handle_color)


func _draw_curve_polygon(curve: Curve, color: Color, side := 1) -> void:
	var points = PoolVector2Array([Vector2(rect_size.x * 0.5, 0)])
	for t in curve.bake_resolution + 1:
		var x := rect_size.x * 0.5 + rect_size.x * 0.5 * side * curve.interpolate_baked(float(t) / curve.bake_resolution)
		if is_equal_approx(x, rect_size.x):
			continue
		points.push_back(Vector2(x, rect_size.y * float(t) / curve.bake_resolution))

	points.push_back(Vector2(rect_size.x * 0.5, rect_size.y))
	draw_colored_polygon(points, color, PoolVector2Array(), null, null, true)


## Returns [distance, index]. None found is [INF, -1]
func find_closest_point(reference_position: Vector2, curve: Curve, closest_distance := INF, radius := 8.0, side := 1) -> Array:
	var closest_index := -1
	for i in curve.get_point_count():
		var position := curve_to_canvas(curve.get_point_position(i), side)
		var distance := (reference_position - position).length_squared()
		if distance < radius * radius and distance < closest_distance:
			closest_index = i
			closest_distance = distance
	if closest_index == -1:
		return [closest_distance, -1]
	return [closest_distance, closest_index, curve]


func find_closest_point_on_both_curves(reference_position: Vector2, radius := 8.0) -> Array:
	var wdi := find_closest_point(reference_position, feedback_effect.weak_right_motor, INF, radius, 1)
	var sdi := find_closest_point(reference_position, feedback_effect.strong_left_motor, wdi[0], radius, -1)
	if sdi[0] < wdi[0]:
		return sdi
	return wdi


func select_closest_point_on_both_curves(reference_position: Vector2, radius := 8.0) -> void:
	var point := find_closest_point_on_both_curves(reference_position, radius)
	if point[1] != -1:
		current_selection[point[2]][point[1]] = point[2].get_point_position(point[1])


func select_points_in_rect(rect: Rect2, curve: Curve, side := 1) -> void:
	assert(rect.abs() == rect)
	for i in range(curve.get_point_count()):
		var position := curve_to_canvas(curve.get_point_position(i), side)
		multiselection_points[curve][i] = rect.has_point(position)


func clear_selection() -> void:
	current_selection.clear()
	current_selection[feedback_effect.weak_right_motor] = {}
	current_selection[feedback_effect.strong_left_motor] = {}
	selection_rect = Rect2()


func selection_empty() -> bool:
	var empty := true
	for curve in current_selection:
		empty = empty and current_selection[curve].empty()
	return empty


func add_point(mouse_position: Vector2) -> int:
	var position := vec2_swap(mouse_position / Vector2(0.5 * rect_size.x, rect_size.y))

	if mouse_position.x < 0.5 * rect_size.x:
		position.y = 1 - position.y
		return feedback_effect.strong_left_motor.add_point(position)
	position.y -= 1
	return feedback_effect.weak_right_motor.add_point(position)


func selection_switch_curve() -> void:
	var new_selection := {feedback_effect.weak_right_motor: {}, feedback_effect.strong_left_motor: {}}
	for curve in current_selection:
		var delete_points := []
		for point in current_selection[curve]:
			var other_curve: Curve= feedback_effect.weak_right_motor if curve == feedback_effect.strong_left_motor else feedback_effect.strong_left_motor
			var idx = other_curve.add_point(curve.get_point_position(point), \
					curve.get_point_left_tangent(point), \
					curve.get_point_right_tangent(point), \
					curve.get_point_left_mode(point), \
					curve.get_point_right_mode(point))

			delete_points.push_back(point)
			new_selection[other_curve][idx] = 0

		delete_points.sort()
		for i in range(delete_points.size(), 0, -1):
			curve.remove_point(delete_points[i - 1])
	current_selection = new_selection


func remove_selection() -> void:
	for curve in current_selection:
		var points: Array = current_selection[curve].keys()
		points.sort()
		for i in range(points.size(), 0, -1):
			curve.remove_point(points[i - 1])
	clear_selection()
	is_selecting = false
	is_moving = false


func calculate_selection_rect() -> void:
	var first := true
	for curve in current_selection:
		for point in current_selection[curve]:
			var position := curve_to_canvas(curve.get_point_position(point), \
					1 if curve == feedback_effect.weak_right_motor else -1)
			if first:
				selection_rect = Rect2(position, Vector2())
				first = false
				continue

			selection_rect = selection_rect.expand(position)


func curve_to_canvas(position: Vector2, side := 1) -> Vector2:
	var t := position.x * rect_size.y
	position.x = position.y * rect_size.x * 0.5 * side + rect_size.x * 0.5
	position.y = t
	return position


static func vec2_swap(vec2: Vector2) -> Vector2:
	var temp := vec2.x
	vec2.x = vec2.y
	vec2.y = temp
	return vec2


static func vec2_clamp(vec2: Vector2, lower := 0.0, upper := 1.0) -> Vector2:
	return Vector2(clamp(vec2.x, lower, upper), clamp(vec2.y, lower, upper))


func calculate_curve_offset(position: Vector2) -> Vector2:
	var offset := (position - selection_start_position) / Vector2(0.5 * rect_size.x, rect_size.y)
	if selection_start_position.x < 0.5 * rect_size.x:
		offset.x *= -1
	return vec2_swap(offset)


func ensure_selection(mouse_position: Vector2) -> void:
	if !selection_rect.has_point(mouse_position):
		clear_selection()
	if selection_empty():
		select_closest_point_on_both_curves(mouse_position)


func start_select(mouse_position: Vector2) -> void:
	selection_rect = Rect2(mouse_position, Vector2())
	multiselection_points = {feedback_effect.weak_right_motor: {}, feedback_effect.strong_left_motor: {}}
	is_selecting = true


func _on_length_changed(value: float) -> void:
	var scale: float = feedback_effect.length / value
	feedback_effect.length = value
	if Input.is_key_pressed(KEY_CONTROL) or !length_scale:
		update()
		return
	for curve in current_selection:
		if scale < 1:
			for point in range(curve.get_point_count()):
				curve.set_point_offset(point, curve.get_point_position(point).x * scale)
		else:
			for i in range(curve.get_point_count(), 0, -1):
				curve.set_point_offset(i - 1, curve.get_point_position(i - 1).x * scale)
	update()


func set_feedback_effect(value: ForceFeedbackEffect) -> void:
	feedback_effect = value
	multiselection_points.clear()
	multiselection_points[feedback_effect.weak_right_motor] = {}
	multiselection_points[feedback_effect.strong_left_motor] = {}
	clear_selection()
	update()
